// CircularList.h

#ifndef __CIRCULARLIST_H__
#define __CIRCULARLIST_H__

#include <assert.h>

#ifndef NULL
#define NULL 0
#endif


template<class T>
class ListNode;

/** represents a doubly-linked circular list. specific items in the list are
 * accessed via ListNode<T> iterators.
 * @see ListNode<T>
 * @author adi karol
 * @version 8/01
 */
template<class T>
class CircularList {
public:
	/// constructor. constructs an empty list.
	CircularList() : head(NULL) {}

	/// copy constructor.
	CircularList(const CircularList<T> &src) : head(NULL) {
		nodeptr head, iter, last;
		if ((head = iter = src.getHead())) {
			last = setHead(head->getItem());
			while ((iter = iter->getNext()) != head)
				last = last->addNext(iter->getItem());
		}
	}

	/// destructor.
	~CircularList() {
		if (head) {
			ListNode<T> *iter = head;
			do {
				ListNode<T> *tmp = iter;
				iter = iter->next;
				delete tmp;
			} while (iter != head);
		}
	}

	/** pointer to a node in the list. serves as iterator for the list.
	 * @see ListNode<T>
	 */
	typedef ListNode<T> *nodeptr;

	/// @return iterator to the head of the list.
	const nodeptr getHead() const { return head; }

	/// @return iterator to the head of the list.
	nodeptr getHead() { return head; }

	/** inserts a first item into the list.
	 * @param item the inserted item
	 * @precondition the list is empty
	 * @return the node of the newly inserted item
	 */
	nodeptr setHead(const T &item) {
		assert(!head);
		head = new ListNode<T>(item);
		return head;
	}

	/** sets a new head. the new head should be a node in the list.
	 * @param _head the new head.
	 */
	void rotateHead(nodeptr _head) {
		head = _head;
	}

protected:
	nodeptr head;

	friend class ListNode<T>;
};


/** represents a node in a doubly-linked list. serves as iterator for the list.
 * @see CircularList<T>
 * @author adi karol
 * @version 8/01
 */
template<class T>
class ListNode {
public:
	/// @return the item associated with the node
	const T &getItem() const { return item; }
	/// @return the item associated with the node
	T &getItem() { return item; }
	/// @return the next node
	const ListNode<T> *getNext() const { return next; }
	/// @return the next node
	ListNode<T> *getNext() { return next; }
	/// @return the previous node
	const ListNode<T> *getPrev() const { return prev; }
	/// @return the previous node
	ListNode<T> *getPrev() { return prev; }
	/** inserts a new item after this node.
	 * @param item the inserted item
	 * @return the node of the newly inserted item
	 */
	ListNode<T> *addNext(const T &item) {
		ListNode<T> *newItem = new ListNode<T>(item, next, this);
		next->prev = newItem;
		return next = newItem;
	}
	/** inserts a new item before this node.
	 * @param item the inserted item
	 * @return the node of the newly inserted item
	 */
	ListNode<T> *addPrev(const T &item) {
		ListNode<T> *newItem = new ListNode<T>(item, this, prev);
		prev->next = newItem;
		return prev = newItem;
	}
	/** deletes the node and removes it from the list.
	 * note: after the call, 'this' is not valid.
	 * @param list the list containing the node
	 * @precondition the node belongs to the list
	 */
	void deleteFrom(CircularList<T> &list) {
		// no check is done to ensure that this node is indeed in the list
		if (list.head->next == list.head) {
			// one item in list
			list.head = NULL;
	} else {
			// more than one item
			next->prev = prev;
			prev->next = next;
			if (list.head == this)
				list.head = next;
		}
		delete this;
	}

protected:
	/** constructors a node.
	 * @param _item the item to be associated with the node
	 * @param _next the next node
	 * @param _prev the previous node
	 */
	ListNode(const T &_item, ListNode<T> *_next, ListNode<T> *_prev) :
		item(_item), next(_next), prev(_prev) { }
	/** constructs a singleton node.
	 * @param _item the item to be associated with the node
	 */
	ListNode(const T &_item) :
		item(_item) { next = prev = this; }

	ListNode<T> *next, *prev;
	T item;

	friend class CircularList<T>;
};


#endif // __CIRCULARLIST_H__

