///////////////////////////////////////////////////////////////////////////
//  TexturedMeshRendering.cpp: implementation of the TexturedMesh class  //
///////////////////////////////////////////////////////////////////////////

#pragma warning( disable : 4786 )
#pragma warning( disable : 4503 )


#include "TexturedMesh.h"

// For OpenGL rendering
// #include <windows.h>            // Window defines

#include <GL/gl.h>              // OpenGL
#include <GL/glu.h>             // GLU library
#include <GL/glut.h>             // GLUT library
#include <GL/glaux.h>			// AUX library
//#include "OpenGLtexture.h"		// for texture handling

//////////////////////////////////////////////////////////////////////
// Rendering functions
//////////////////////////////////////////////////////////////////////

/******************************************************************/
// Render2DMeshToBMP:
//
// Renders the mesh to a buffer and saves in a file
// called filename in Windows Bitmap format.
// If render_texture is true and the flat mesh has
// texture, the function renders with the texture.
// If render_texture is false, or if the mesh doesn't
// have a texture, the function renders wireframe.
// Works only for 2D mesh. If m_is2Dmesh is false or
// if the function fails for some reason, it returns false,
// returns true on success.
/******************************************************************/
#ifdef MY_IGNORE
bool TexturedMesh::Render2DMeshToBMP(const char *filename, bool render_texture) const {
	if ( /*(m_is2Dmesh == false) || */
		(render_texture && (m_uv_per_triangle.size() == 0)) )
		return false;
	
	
	// need_texture says weather we NEED to render texture.
	// Even if render_texture is true, if the mesh doesn't
	// have texture we render wireframe...
	bool need_texture = render_texture && (m_uv_per_triangle.size() > 0);
	
	if (need_texture) {
		// Loading texture for OpenGL rendering
		LoadTexture(m_textures[0], false, false);
	}
	
	// Determine the dimensions of the bitmap:
	Box box = GetBoundingBox();
	UINT width = (box.GetRight() - box.GetLeft());
	UINT height = (box.GetFront() - box.GetBack());
	
	// For translation so that all coordinates are positive;
	float tx = -1 * box.GetLeft(); 
	float ty = -1 * box.GetBack();
	
	width = max(width, 1);
	height = max(height, 1);
	
	// Creating the bitmap and the rendering context
	
	BMPStruct bmpStruct;
	
	bmpStruct.width = width;
	bmpStruct.height = height;
	
	HDC                   dc;     /* Device context */
	PIXELFORMATDESCRIPTOR pfd;    /* Pixel format description structure */
	int                   pf;     /* Pixel format */
	HGLRC                 rc;     /* OpenGL rendering context */
	
	/* Create the device context... */
	dc = CreateCompatibleDC(NULL);
	
	/* Create and bind the bitmap */
	memset(&(bmpStruct.info), 0, sizeof(bmpStruct.info));
	bmpStruct.info.bmiHeader.biSize        = sizeof(BITMAPINFOHEADER);
	bmpStruct.info.bmiHeader.biWidth       = bmpStruct.width;
	bmpStruct.info.bmiHeader.biHeight      = bmpStruct.height;
	bmpStruct.info.bmiHeader.biPlanes      = 1;
	bmpStruct.info.bmiHeader.biBitCount    = 24;
	bmpStruct.info.bmiHeader.biCompression = BI_RGB;
	
	bmpStruct.bitmap = CreateDIBSection(dc,&(bmpStruct.info),DIB_RGB_COLORS, (void**)&(bmpStruct.bits),NULL,0);
	
	SelectObject(dc, bmpStruct.bitmap);
	
	/* Set the pixel format... */
	memset(&pfd, 0, sizeof(pfd));
	pfd.nSize      = sizeof(pfd);
	pfd.nVersion   = 1;
	pfd.dwFlags    = PFD_DRAW_TO_BITMAP | PFD_SUPPORT_OPENGL;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 24;
	pfd.cRedBits   = 8;
	pfd.cGreenBits = 8;
	pfd.cBlueBits  = 8;
	pfd.cDepthBits = 24;
	
	pf = ChoosePixelFormat(dc, &pfd);
	SetPixelFormat(dc, pf, &pfd);
	
	/* Create and bind the rendering context */
	rc = wglCreateContext(dc);
	wglMakeCurrent(dc, rc);
	
	
	// Rendering the mesh!
	
	if (need_texture) {
		glEnable(GL_TEXTURE_2D);
		// No filter - to achieve fast rendering
		SetTextureFilter(TF_NONE);
		// In order to make the textures look good 
		glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
	}
	else {// rendering wireframe 
		glDisable(GL_CULL_FACE);
		
		glDisable(GL_TEXTURE_2D);
		glEnable(GL_POINT_SMOOTH);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glColor3f(0.3f,0.3f,0.3f);
	}
	
	// Reset the viewport
	glViewport(0, 0, width, height);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	glOrtho(0.0f, width, 0.0f, height, 1.0f, -1.0f);
	// Set clear color to White
	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
	
	// Clear the window
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	glDisable(GL_LIGHTING);
	
	
	//	Sending the mesh to OpenGL pipeline 
	glBegin(GL_TRIANGLES);
	
	for(UINT trippletIndex = 0; trippletIndex < m_faces.size(); ++ trippletIndex) {		
		const Tripplet & tripplet = m_faces[trippletIndex];
		UINT  firstPointIndex = tripplet.GetFirstIndex();
		UINT  secondPointIndex = tripplet.GetSecondIndex();
		UINT  thirdPointIndex = tripplet.GetThirdIndex();
		
		const Point3D & firstPoint = m_vertices[firstPointIndex];
		const Point3D & secondPoint = m_vertices[secondPointIndex];
		const Point3D & thirdPoint = m_vertices[thirdPointIndex];
		
		if (need_texture) {
			const Tripplet & uv_tripplet = m_uv_per_triangle[trippletIndex];
			
			const Point2D & firstTexel = m_uv_coordinates[uv_tripplet.GetFirstIndex()];
			const Point2D & secondTexel = m_uv_coordinates[uv_tripplet.GetSecondIndex()];
			const Point2D & thirdTexel = m_uv_coordinates[uv_tripplet.GetThirdIndex()];
			
			glTexCoord2f((float)firstTexel.GetX(), (float)firstTexel.GetY());
			glVertex2f((float)firstPoint.GetX()+tx,(float)firstPoint.GetY()+ty);
			glTexCoord2f((float)secondTexel.GetX(), (float)secondTexel.GetY());
			glVertex2f((float)secondPoint.GetX()+tx,(float)secondPoint.GetY()+ty);
			glTexCoord2f(thirdTexel.GetX(), (float)thirdTexel.GetY());
			glVertex2f((float)thirdPoint.GetX()+tx,(float)thirdPoint.GetY()+ty);
		}
		else {
			glVertex2f((float)firstPoint.GetX()+tx,(float)firstPoint.GetY()+ty);
			glVertex2f((float)secondPoint.GetX()+tx,(float)secondPoint.GetY()+ty);
			glVertex2f((float)thirdPoint.GetX()+tx,(float)thirdPoint.GetY()+ty);
		}
	}
	glEnd();
	glFlush();
	
	
	/* Freeing all unneccery data */
	wglDeleteContext(rc);
	DeleteDC(dc);
	
	int bSaveSuccess = SaveBitmapTexture(filename, bmpStruct);
	
	return (bSaveSuccess != 0);
}
#endif // IGNORE


// renders the mesh using OpenGL
void TexturedMesh::RenderTriangles(bool normal_per_vertex, 
								   int color_coding, 
								   const unsigned char *color) const {

	int n = m_vertices.size();
	bool anything2draw = true;

	if (normal_per_vertex) {
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);

		anything2draw = (!m_vertices.empty() && !m_vertex_normals.empty());
		glVertexPointer(3, GL_FLOAT, 0, m_vertices.begin());
		if (m_use_own_normals && !m_own_vertex_normals.empty()) {
			glNormalPointer(GL_FLOAT, 0, (float *)m_own_vertex_normals.begin());
		}
		else {
			glNormalPointer(GL_FLOAT, 0, (float *)m_vertex_normals.begin());
		}

		switch (color_coding) {
		case 0:
			if (!m_color_per_vertex.empty()) {
				glEnableClientState(GL_COLOR_ARRAY);
				glColorPointer(3, GL_FLOAT, 0, m_color_per_vertex.begin());
			}
			else {
				glDisableClientState(GL_COLOR_ARRAY);
			}
			break;
		case 1:
			if (!m_color_per_vertex1.empty()) {
				glEnableClientState(GL_COLOR_ARRAY);
				glColorPointer(3, GL_FLOAT, 0, m_color_per_vertex1.begin());
			}
			else {
				glDisableClientState(GL_COLOR_ARRAY);
			}
			break;
		default:
			glDisableClientState(GL_COLOR_ARRAY);
			break;
		}
		
		if (color_coding > 1) {
			if (color != NULL)
				glColor3ubv(color);
		}

		if (anything2draw)
			glDrawElements(GL_TRIANGLES, 3*m_faces.size(), GL_UNSIGNED_INT, m_faces.begin());
	}
	else { // for flat shading
		int ind;

		const std::vector<Point3D> * v;
		const std::vector<Point3D> * nf;

		v = &m_vertices;
		nf = &m_face_normals;

		if (v->empty() || nf->empty())
			return;

		if (color != NULL)
			glColor3ubv(color);

		glBegin(GL_TRIANGLES);
		for(int i = 0 ; i < m_faces.size(); ++i) {
			ind = m_faces[i].GetFirstIndex();
			
			glNormal3fv((*nf)[i].GetCoordArray());
			
			glVertex3fv((*v)[ind].GetCoordArray());	
			ind = m_faces[i].GetSecondIndex();
			glVertex3fv((*v)[ind].GetCoordArray());
			ind = m_faces[i].GetThirdIndex();
			glVertex3fv((*v)[ind].GetCoordArray());
		}
		glEnd();
	}
}

// renders the mesh using OpenGL
void TexturedMesh::RenderPoints() const {
	int num;

	glEnable(GL_POINT_SMOOTH);

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);

	glVertexPointer(3, GL_FLOAT, 0, m_vertices.begin());
	glNormalPointer(GL_FLOAT, 0, (float *)m_vertex_normals.begin());
	num = m_vertices.size();
		
	glColorPointer(3, GL_FLOAT, 0, m_color_per_vertex1.begin());
	
	glDrawArrays(GL_POINTS, 0, num);
}


// if render_all is false, renders only boundary vertices
void TexturedMesh::RenderVertexBalls(float radius, 
									 int render_all, 
									 int num_segments_width,
									 int num_segments_height) {
	// Initializing mesh auxiliary structures
	if (m_edge_list.empty())
		CreateEdgeList();
	if (m_edge_to_triangles_map.empty())
		CreateEdgeToTrianglesMap();
	if (m_triangles_adjacency_list.empty())
		CreateTrianglesAdjacencyList();
	
	std::vector<Point3D>::const_iterator v;

	v = m_vertices.begin();
	
	for (int i = 0; i < m_vertices.size(); ++i) {
		if ((render_all==0) ||
			(IsBoundaryVertex(i) && render_all == 1) ||
			(IsIsolatedVertex(i) && render_all == 2) ||
			(IsHangingVertex(i)  && render_all == 3) ) {
			glTranslatef(v[i].GetX(), v[i].GetY(), v[i].GetZ());
			glutSolidSphere(radius, /* radius */ 
				num_segments_width,     /* slices */ 
				num_segments_height      /* stacks */ );
			glTranslatef(-v[i].GetX(), -v[i].GetY(), -v[i].GetZ());
		}
	}
}
